/*******************************************************************************
  MRF24WG Private Configuration Support

  File Name:
    drv_wifi_iwpriv.c

  Summary:
    Configure optional (private) parameters of MRF24WG driver.

  Description:
    Functions in this module support the connection process for the
    MRF24WG.
 *******************************************************************************/

//DOM-IGNORE-BEGIN
/*******************************************************************************
Copyright (c) 2013 released Microchip Technology Inc.  All rights reserved.

Microchip licenses to you the right to use, modify, copy and distribute
Software only when embedded on a Microchip microcontroller or digital signal
controller that is integrated into your product or third party product
(pursuant to the sublicense terms in the accompanying license agreement).

You should refer to the license agreement accompanying this Software for
additional information regarding your rights and obligations.

SOFTWARE AND DOCUMENTATION ARE PROVIDED AS IS WITHOUT WARRANTY OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION, ANY WARRANTY OF
MERCHANTABILITY, TITLE, NON-INFRINGEMENT AND FITNESS FOR A PARTICULAR PURPOSE.
IN NO EVENT SHALL MICROCHIP OR ITS LICENSORS BE LIABLE OR OBLIGATED UNDER
CONTRACT, NEGLIGENCE, STRICT LIABILITY, CONTRIBUTION, BREACH OF WARRANTY, OR
OTHER LEGAL EQUITABLE THEORY ANY DIRECT OR INDIRECT DAMAGES OR EXPENSES
INCLUDING BUT NOT LIMITED TO ANY INCIDENTAL, SPECIAL, INDIRECT, PUNITIVE OR
CONSEQUENTIAL DAMAGES, LOST PROFITS OR LOST DATA, COST OF PROCUREMENT OF
SUBSTITUTE GOODS, TECHNOLOGY, SERVICES, OR ANY CLAIMS BY THIRD PARTIES
(INCLUDING BUT NOT LIMITED TO ANY DEFENSE THEREOF), OR OTHER SIMILAR COSTS.
 *******************************************************************************/

/******************/
/*    INCLUDES    */
/******************/
#include "tcpip/src/tcpip_private.h"
#include "drv_wifi_iwpriv.h"

/*******************/
/*    VARIABLES    */
/*******************/
extern DRV_WIFI_SCAN_STATUS g_scanStatus;
extern DRV_WIFI_CONFIG_DATA *p_wifi_configData;

static uint8_t s_prescan = false;
static uint8_t s_prescan_waiting = false; // WF_PRESCAN - This is used only to allow prescan once.

/*******************/
/*    FUNCTIONS    */
/*******************/
void iwpriv_prescan_start(void)
{
    s_prescan_waiting = true;
    g_drv_wifi_priv.isScanDone = false;
    DRV_WIFI_HostScanStart();
}

bool iwpriv_prescan_done(void)
{
    if (s_prescan_waiting && g_drv_wifi_priv.isScanDone) {
        s_prescan_waiting = false;
        return true;
    }

    return false;
}

void iwpriv_prescan_set(bool scan)
{
    if (scan) {
        s_prescan = true;
        s_prescan_waiting = false;
    } else {
        s_prescan = false;
        s_prescan_waiting = false;
    }
}

bool iwpriv_prescan_get(void)
{
    return s_prescan;
}

void iwpriv_scan_start(void)
{
    uint8_t netType;

    iwpriv_nettype_get(&netType);
    if (netType == DRV_WIFI_NETWORK_TYPE_SOFT_AP && isLinkUp()) {
        SYS_CONSOLE_MESSAGE("MRF24WG cannot perform scan in Soft AP mode ... \r\n");
    } else if (DRV_WIFI_ScanStart() == DRV_WIFI_SUCCESS) {
        SYS_CONSOLE_MESSAGE("Scan started ... \r\n");
        SCAN_SET_DISPLAY(g_scanStatus.scanState);
        g_scanStatus.displayIdx = 0;
    }
}

IWPRIV_SCAN_RESULT iwpriv_scanstatus_get(void)
{
    IWPRIV_SCAN_RESULT ret;

    if(IS_SCAN_IN_PROGRESS(g_scanStatus.scanState)) {
        SYS_CONSOLE_MESSAGE((const char *)"Scan in progress.\r\n ");
        ret = IWPRIV_SCAN_IN_PROGRESS;
    } else if(IS_SCAN_STATE_DISPLAY(g_scanStatus.scanState) && (g_scanStatus.numScanResults > 0)){
        ret = IWPRIV_SCAN_SUCCEEDED;
    } else if (IS_SCAN_STATE_DISPLAY(g_scanStatus.scanState) && (g_scanStatus.numScanResults == 0)) {
        ret = IWPRIV_SCAN_NO_APS_FOUND;
    } else {
        ret = IWPRIV_SCAN_IDLE;
    }

    return ret;
}

uint16_t iwpriv_numberofscanresults_get(void)
{
    return g_scanStatus.numScanResults;
}

void *iwpriv_scanresult_get(uint8_t index)
{
    if (index <= iwpriv_numberofscanresults_get())
        return &g_scanResults[index];
    else
        return NULL;
}

IWPRIV_STATUS iwpriv_scanresults_display(void)
{
    if (IS_SCAN_STATE_DISPLAY(g_scanStatus.scanState)) {
        DRV_WIFI_ScanResultDisplayManager();
        return IWPRIV_IN_PROGRESS;
    }

    return IWPRIV_READY;
}

IWPRIV_CONN_STATUS iwpriv_connstatus_get(void)
{
    if (isLinkUp()) {
        if (g_drv_wifi_priv.isConnReestablished) {
            g_drv_wifi_priv.isConnReestablished = false;
            return IWPRIV_CONNECTION_REESTABLISHED;
        } else {
            return IWPRIV_CONNECTION_SUCCESSFUL;
        }
    } else {
        if (g_drv_wifi_priv.isConnPermanentlyLost) {
            return IWPRIV_CONNECTION_FAILED;
        } else {
            return IWPRIV_CONNECTION_IDLE;
        }
    }

    return IWPRIV_CONNECTION_IN_PROGRESS;
}

void iwpriv_config_read(DRV_WIFI_CONFIG_DATA *wifi_config)
{
    memcpy(wifi_config, p_wifi_configData, sizeof(DRV_WIFI_CONFIG_DATA));
}

void iwpriv_config_write(void *wifi_config)
{
    memcpy(p_wifi_configData, wifi_config, sizeof(DRV_WIFI_CONFIG_DATA));
    DRV_WIFI_ConfigDataSave();
}

bool iwpriv_is_servermode(void)
{
    uint8_t netType;

    iwpriv_nettype_get(&netType);
    if (netType == DRV_WIFI_NETWORK_TYPE_SOFT_AP || netType == DRV_WIFI_NETWORK_TYPE_ADHOC)
        return true;
    else
        return false;
}

void iwpriv_devinfo_get(void *info)
{
    DRV_WIFI_DeviceInfoGet((DRV_WIFI_DEVICE_INFO *)info);
}

void iwpriv_ssid_set(uint8_t *ssid, uint8_t ssidLen)
{
    DRV_WIFI_SsidSet(ssid, ssidLen);
}

void iwpriv_ssid_get(uint8_t *ssid, uint8_t *ssidLen)
{
    DRV_WIFI_SsidGet(ssid, ssidLen);
}

void iwpriv_nettype_set(uint8_t netType)
{
    DRV_WIFI_NetworkTypeSet(netType);
}

void iwpriv_nettype_get(uint8_t *netType)
{
    DRV_WIFI_NetworkTypeGet(netType);
}

void iwpriv_adhocctx_set(DRV_WIFI_ADHOC_NETWORK_CONTEXT *p_context)
{
    DRV_WIFI_AdhocContextSet(p_context);
}

IWPRIV_STATUS iwpriv_mcastfilter_set(uint8_t *addr)
{
    if (MRF24W_MACMulticastFilterSet(NULL, (TCPIP_MAC_ADDR *)addr) == TCPIP_MAC_RES_OK)
        return IWPRIV_READY;

    return IWPRIV_ERROR;
}

void iwpriv_powersave_config(bool enabled)
{
    if (enabled) {
        DRV_WIFI_PS_POLL_CONTEXT psPollContext;
        psPollContext.dtimInterval = DRV_WIFI_DEFAULT_PS_DTIM_INTERVAL;
        psPollContext.listenInterval = DRV_WIFI_DEFAULT_PS_LISTEN_INTERVAL;
        psPollContext.useDtim = true;

        DRV_WIFI_PsPollEnable(&psPollContext);
    } else {
        DRV_WIFI_PsPollDisable();
    }
}

void iwpriv_initialconn_set(bool connect)
{
    g_drv_wifi_priv.initConn = connect;
}

void iwpriv_leftclient_get(bool *updated, TCPIP_MAC_ADDR *addr)
{
    /* TODO: Implement DHCP server sync function. */
    *updated = false;
}

uint8_t iwpriv_initstatus_get(void)
{
    if (g_drv_wifi_priv.isDriverOpen)
        return IWPRIV_READY;
    else
        return IWPRIV_IN_PROGRESS;
}

//DOM-IGNORE-END
