/*******************************************************************************
  MRF24WG Private Configuration Support

  Company:
    Microchip Technology Inc.
  
  File Name:
    drv_wifi_iwpriv.h

  Summary:
    Configure optional (private) parameters of MRF24WG driver.

  Description:
    Functions in this module support the connection process for the
    MRF24WG.
 *******************************************************************************/

// DOM-IGNORE-BEGIN
/*******************************************************************************
Copyright (c) 2015 released Microchip Technology Inc. All rights reserved.

Microchip licenses to you the right to use, modify, copy and distribute
Software only when embedded on a Microchip microcontroller or digital signal
controller that is integrated into your product or third party product
(pursuant to the sublicense terms in the accompanying license agreement).

You should refer to the license agreement accompanying this Software for
additional information regarding your rights and obligations.

SOFTWARE AND DOCUMENTATION ARE PROVIDED AS IS WITHOUT WARRANTY OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION, ANY WARRANTY OF
MERCHANTABILITY, TITLE, NON-INFRINGEMENT AND FITNESS FOR A PARTICULAR PURPOSE.
IN NO EVENT SHALL MICROCHIP OR ITS LICENSORS BE LIABLE OR OBLIGATED UNDER
CONTRACT, NEGLIGENCE, STRICT LIABILITY, CONTRIBUTION, BREACH OF WARRANTY, OR
OTHER LEGAL EQUITABLE THEORY ANY DIRECT OR INDIRECT DAMAGES OR EXPENSES
INCLUDING BUT NOT LIMITED TO ANY INCIDENTAL, SPECIAL, INDIRECT, PUNITIVE OR
CONSEQUENTIAL DAMAGES, LOST PROFITS OR LOST DATA, COST OF PROCUREMENT OF
SUBSTITUTE GOODS, TECHNOLOGY, SERVICES, OR ANY CLAIMS BY THIRD PARTIES
(INCLUDING BUT NOT LIMITED TO ANY DEFENSE THEREOF), OR OTHER SIMILAR COSTS.
 *******************************************************************************/
// DOM-IGNORE-END

#ifndef _DRV_WIFI_IWPRIV_H
#define _DRV_WIFI_IWPRIV_H

#include "driver/wifi/mrf24w/src/drv_wifi_config_data.h"

typedef enum {
    IWPRIV_SCAN_SUCCEEDED,
    IWPRIV_SCAN_IN_PROGRESS,
    IWPRIV_SCAN_IDLE,
    IWPRIV_SCAN_NO_APS_FOUND
} IWPRIV_SCAN_RESULT;

typedef enum {
    IWPRIV_ERROR = -1,
    IWPRIV_READY,
    IWPRIV_IN_PROGRESS
} IWPRIV_STATUS;

typedef enum {
    IWPRIV_CONNECTION_FAILED = -1,
    IWPRIV_CONNECTION_IDLE,
    IWPRIV_CONNECTION_SUCCESSFUL,
    IWPRIV_CONNECTION_IN_PROGRESS,
    IWPRIV_CONNECTION_REESTABLISHED
} IWPRIV_CONN_STATUS;

//*******************************************************************************
/*
  Function:
        void iwpriv_prescan_start(void)

  Summary:
    Starts prescan.
    <p><b>Implementation:</b> Dynamic</p>

  Description:
    This function directs the Wi-Fi driver to start a prescan.

  Precondition:
    Wi-Fi initialization must be complete.

  Returns:
    None.

  Remarks:
    None.
 */
void iwpriv_prescan_start(void);

//*******************************************************************************
/*
  Function:
        bool iwpriv_prescan_done(void)

  Summary:
    Checks if the prescan is complete.
    <p><b>Implementation:</b> Dynamic</p>

  Description:
    This function checks if the prescan is complete.

  Precondition:
    Wi-Fi initialization must be complete.

  Returns:
    None.

  Remarks:
    None.
 */
bool iwpriv_prescan_done(void);

//*******************************************************************************
/*
  Function:
        void iwpriv_prescan_set(bool scan)

  Summary:
    To run prescan or not.
    <p><b>Implementation:</b> Dynamic</p>

  Description:
    This function controls whether or not to run prescan.

  Precondition:
    Wi-Fi initialization must be complete.

  Parameters:
    scan -  true: run prescan before next connection
           false: do not run prescan before next connection

  Returns:
    None.

  Remarks:
    Prescan means the scan runs before the module is connected. It needs
    to use multiple functions in this file. Please refer to the Easy Configuration
    demonstration to see the correct usage of prescan.

    After the the module is connected, the MRF24WG can also do regular scans in
    Infrastructure mode or Ad-Hoc mode, but it cannot perform a scan in Soft AP
    mode. It also cannot perform a scan when the connection is in progress.
 */
void iwpriv_prescan_set(bool scan);

//*******************************************************************************
/*
  Function:
        bool iwpriv_prescan_get(void)

  Summary:
    To see if prescan will run before next connection.
    <p><b>Implementation:</b> Dynamic</p>

  Description:
    This function checks whether or not the prescan will run before next connection.

  Precondition:
    Wi-Fi initialization must be complete.

  Returns:
    None.

  Remarks:
    None.
 */
bool iwpriv_prescan_get(void);

//*******************************************************************************
/*
  Function:
        void iwpriv_scan_start(void)

  Summary:
    Starts scan.
    <p><b>Implementation:</b> Dynamic</p>

  Description:
    The function starts a Wi-Fi scan.

  Precondition:
    Wi-Fi initialization must be complete.

  Returns:
    None.

  Remarks:
    None.
 */
void iwpriv_scan_start(void);

//*******************************************************************************
/*
  Function:
        IWPRIV_SCAN_RESULT iwpriv_scanstatus_get(void)

  Summary:
    Gets the prescan status.
    <p><b>Implementation:</b> Dynamic</p>

  Description:
    This function gets the prescan status.

  Precondition:
    Wi-Fi initialization must be complete.

  Returns:
    Prescan status: IWPRIV_SCAN_IDLE, IWPRIV_SCAN_IN_PROGRESS,
    IWPRIV_SCAN_NO_APS_FOUND or IWPRIV_SCAN_SUCCEEDED. See the definition for the
    IWPRIV_SCAN_RESULT structure.

  Remarks:
    None.
 */
IWPRIV_SCAN_RESULT iwpriv_scanstatus_get(void);

//*******************************************************************************
/*
  Function:
        uint16_t iwpriv_numberofscanresults_get(void)

  Summary:
    Gets the number of scan results.
    <p><b>Implementation:</b> Dynamic</p>

  Description:
    This function gets the number of scan results.

  Precondition:
    Wi-Fi initialization must be complete.

  Returns:
    Number of scan results.

  Remarks:
    None.
 */
uint16_t iwpriv_numberofscanresults_get(void);

//*******************************************************************************
/*
  Function:
        void *iwpriv_scanresult_get(uint8_t index)

  Summary:
    Gets one scan result.
    <p><b>Implementation:</b> Dynamic</p>

  Description:
    This function gets one scan result, which includes the detailed information
    of an AP.

  Precondition:
    Wi-Fi initialization must be complete.

  Parameters:
    index - index of the scan result

  Returns:
    Pointer to where the scan result is written. It uses the g_scanResults[]
    array to store results, which eliminates the need to allocate additional memory.

  Remarks:
    None.
 */
void *iwpriv_scanresult_get(uint8_t index);

//*******************************************************************************
/*
  Function:
        IWPRIV_STATUS iwpriv_scanresults_display(void)

  Summary:
    Outputs all of the scan results.
    <p><b>Implementation:</b> Dynamic</p>

  Description:
    This function outputs all of the scan results to the console terminal.

  Precondition:
    Wi-Fi initialization must be complete.

  Returns:
    Status of the output operation, IWPRIV_IN_PROGRESS or IWPRIV_READY. See
    definition for the IWPRIV_STATUS structure.

  Remarks:
    None.
 */
IWPRIV_STATUS iwpriv_scanresults_display(void);

//*******************************************************************************
/*
  Function:
        IWPRIV_CONN_STATUS iwpriv_connstatus_get(void)

  Summary:
    Gets the Wi-Fi connection status.
    <p><b>Implementation:</b> Dynamic</p>

  Description:
    This function gets the Wi-Fi connection status.

  Precondition:
    Wi-Fi initialization must be complete.

  Returns:
    Status of current Wi-Fi connection. See the definition for the
    IWPRIV_CONN_STATUS structure.

  Remarks:
    IWPRIV_CONNECTION_FAILED does not necessarily mean
    DRV_WIFI_EVENT_CONNECTION_FAILED. It stands on the application's
    perspective, and actually can be customized. For example, in the Web Server
    demonstration's use case, DRV_WIFI_EVENT_CONNECTION_PERMANENTLY_LOST is treated as a
    fail case and will trigger the application to restart.
 */
IWPRIV_CONN_STATUS iwpriv_connstatus_get(void);

//*******************************************************************************
/*
  Function:
        void iwpriv_config_read(DRV_WIFI_CONFIG_DATA *wifi_config)

  Summary:
    Reads the Wi-Fi context configuration.
    <p><b>Implementation:</b> Dynamic</p>

  Description:
    This function reads the current Wi-Fi context configuration, copies and stores
    the whole structure to the pointer passed to the function.

  Precondition:
    Wi-Fi initialization must be complete.

  Parameters:
    wifi_config - pointer to where the context configuration is written

  Returns:
    None.

  Remarks:
    None.
 */
void iwpriv_config_read(DRV_WIFI_CONFIG_DATA *wifi_config);

//*******************************************************************************
/*
  Function:
        void iwpriv_config_write(void *wifi_config)

  Summary:
    Writes to the Wi-Fi context configuration which is currently used by Wi-Fi
    driver.
    <p><b>Implementation:</b> Dynamic</p>

  Description:
    This function reads from a passed pointer, copies everything from it, and
    writes to the Wi-Fi context configuration, which is currently used by the Wi-Fi
    driver.

  Precondition:
    Wi-Fi initialization must be complete.

  Parameters:
    wifi_config - pointer to where the context configuration is stored

  Returns:
    None.

  Remarks:
    None.
 *******************************************************************************/
void iwpriv_config_write(void *wifi_config);

//*******************************************************************************
/*
  Function:
        bool iwpriv_is_servermode(void)

  Summary:
    Checks if the passed Wi-Fi context configuration is operating in server
    mode.
    <p><b>Implementation:</b> Dynamic</p>

  Description:
    This function checks if the passed Wi-Fi context configuration is operating
    in server mode, which includes Ad-Hoc mode and Soft AP mode.

  Precondition:
    Wi-Fi initialization must be complete.

  Returns:
    - true  - Wi-Fi context configuration is operating in server mode
    - false - Wi-Fi context configuration is not operating in server mode

  Remarks:
    None.
 */
bool iwpriv_is_servermode(void);

//*******************************************************************************
/*
  Function:
        void iwpriv_devinfo_get(void *info)

  Summary:
    Gets the device information.
    <p><b>Implementation:</b> Dynamic</p>

  Description:
    This function returns the device information, which is either MRF24WG or
    MRF24WN.

  Precondition:
    Wi-Fi initialization must be complete.

  Parameters:
    info - pointer to where the device information is written

  Returns:
    None.

  Remarks:
    This function is different from the function DRV_WIFI_DeviceInfoGet.
 */
void iwpriv_devinfo_get(void *info);

//*******************************************************************************
/*
  Function:
        void iwpriv_ssid_set(uint8_t *ssid, uint8_t ssidLen)

  Summary:
    Sets the current SSID.
    <p><b>Implementation:</b> Dynamic</p>

  Description:
    This function sets the current SSID.

  Precondition:
    Wi-Fi initialization must be complete.

  Parameters:
    ssid - pointer to where the SSID is stored
    ssidLen - pointer to where the SSID length is stored

  Returns:
    None.

  Remarks:
    None.
 */
void iwpriv_ssid_set(uint8_t *ssid, uint8_t ssidLen);

//*******************************************************************************
/*
  Function:
        void iwpriv_ssid_get(uint8_t *ssid, uint8_t *ssidLen)

  Summary:
    Gets the current SSID.
    <p><b>Implementation:</b> Dynamic</p>

  Description:
    This function returns the current SSID.

  Precondition:
    Wi-Fi initialization must be complete.

  Parameters:
    ssid - pointer to where the SSID is written
    ssidLen - pointer to where the SSID length is written

  Returns:
    None.

  Remarks:
    None.
 */
void iwpriv_ssid_get(uint8_t *ssid, uint8_t *ssidLen);

//*******************************************************************************
/*
  Function:
        void iwpriv_nettype_set(uint8_t netType)

  Summary:
    Sets the current network type.
    <p><b>Implementation:</b> Dynamic</p>

  Description:
    This function sets the current network type.

  Precondition:
    Wi-Fi initialization must be complete.

  Parameters:
    netType - the network type to set

  Returns:
    None.

  Remarks:
    None.
 */
void iwpriv_nettype_set(uint8_t netType);

//*******************************************************************************
/*
  Function:
        void iwpriv_nettype_get(uint8_t *netType)

  Summary:
    Gets the current network type.
    <p><b>Implementation:</b> Dynamic</p>

  Description:
    This function returns the current network type.

  Precondition:
    Wi-Fi initialization must be complete.

  Parameters:
    netType - pointer to where the network type is written

  Returns:
    None.

  Remarks:
    None.
 */
void iwpriv_nettype_get(uint8_t *netType);

//*******************************************************************************
/*
  Function:
        void iwpriv_adhocctx_set(DRV_WIFI_ADHOC_NETWORK_CONTEXT *p_context)

  Summary:
    Sets the Ad-Hoc network context information.
    <p><b>Implementation:</b> Dynamic</p>

  Description:
    This function sets the current Ad-Hoc network context information by reading
    from a passed pointer.

  Precondition:
    Wi-Fi initialization must be complete.

  Parameters:
    p_context - pointer to where the Ad-Hoc network context is stored

  Returns:
    None.

  Remarks:
    None.
 */
void iwpriv_adhocctx_set(DRV_WIFI_ADHOC_NETWORK_CONTEXT *p_context);

//*******************************************************************************
/*
  Function:
        IWPRIV_STATUS iwpriv_mcastfilter_set(uint8_t *addr)

  Summary:
    Adds a MAC address to the multi-cast filter.
    <p><b>Implementation:</b> Dynamic</p>

  Description:
    This function adds a MAC address to the multi-cast filter.

  Precondition:
    Wi-Fi initialization must be complete.

  Parameters:
    addr - pointer to where the MAC address is stored

  Returns:
    Status of the set operation, IWPRIV_READY or IWPRIV_ERROR. See definition for
    the IWPRIV_STATUS structure.

  Remarks:
    None.
 */
IWPRIV_STATUS iwpriv_mcastfilter_set(uint8_t *addr);

//*******************************************************************************
/*
  Function:
        void iwpriv_powersave_config(bool enabled)

  Summary:
    Enables or disables Power Save mode in Wi-Fi driver.
    <p><b>Implementation:</b> Dynamic</p>

  Description:
    This function enables or disables Power Save mode in Wi-Fi driver, which depends on
    the passed boolean value. 

  Precondition:
    Wi-Fi initialization must be complete.

  Parameters:
    enabled - boolean value which indicates to enable or disable Power Save mode
              in Wi-Fi driver

  Returns:
    None.

  Remarks:
    None.
 */
void iwpriv_powersave_config(bool enabled);

//*******************************************************************************
/*
  Function:
        void iwpriv_initialconn_set(bool connect)

  Summary:
    Sets the initial connection status of Wi-Fi driver.
    <p><b>Implementation:</b> Dynamic</p>

  Description:
    This function sets the initial connection status of Wi-Fi driver. After
    Wi-Fi initialization, it decides whether or not to start the Wi-Fi connection.

  Precondition:
    Wi-Fi initialization must be complete.

  Parameters:
    connect - boolean value which indicates whether or not to start an initial connect

  Returns:
    None.

  Remarks:
    This function is mainly used to implement prescan. It has to be called
    before Wi-Fi driver's initialization is finished to be effective.
 */
void iwpriv_initialconn_set(bool connect);

//*******************************************************************************
/*
  Function:
        void iwpriv_leftclient_get(bool *updated, TCPIP_MAC_ADDR *addr)

  Summary:
    Gets the left client's information.
    <p><b>Implementation:</b> Dynamic</p>

  Description:
    This function returns the left client's information when the Wi-Fi module works
    in server mode and has the DHCP Server enabled.

  Precondition:
    Wi-Fi initialization must be complete.

  Parameters:
    updated - if the left client's information needs to be updated
    addr - MAC address of the left client

  Returns:
    None.

  Remarks:
    None.
 */
void iwpriv_leftclient_get(bool *updated, TCPIP_MAC_ADDR *addr);

//*******************************************************************************
/*
  Function:
        uint8_t iwpriv_initstatus_get(void)

  Summary:
    Gets the initialization status of Wi-Fi driver.
    <p><b>Implementation:</b> Dynamic</p>

  Description:
    This function returns the initialization status of the Wi-Fi driver.

  Precondition:
    None.

  Returns:
    Current initialization status of the Wi-Fi driver (IWPRIV_READY or IWPRIV_IN_PROGRESS).

  Remarks:
    None.
*/
uint8_t iwpriv_initstatus_get(void);

#endif /* _DRV_WIFI_IWPRIV_H */
