/*******************************************************************************
  MRF24WG Wi-Fi Driver Scan Helper Implementation

  File Name:
    drv_wifi_scan_helper.c

  Summary:
    MRF24WG Wi-Fi Driver Scan Helper Implementation

  Description:
    -Provides access to MRF24WG WiFi controller
    -Reference: MRF24WG Datasheet, IEEE 802.11 Standard
 *******************************************************************************/

//DOM-IGNORE-BEGIN
/*******************************************************************************
Copyright (c) 2013 released Microchip Technology Inc.  All rights reserved.

Microchip licenses to you the right to use, modify, copy and distribute
Software only when embedded on a Microchip microcontroller or digital signal
controller that is integrated into your product or third party product
(pursuant to the sublicense terms in the accompanying license agreement).

You should refer to the license agreement accompanying this Software for
additional information regarding your rights and obligations.

SOFTWARE AND DOCUMENTATION ARE PROVIDED AS IS WITHOUT WARRANTY OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION, ANY WARRANTY OF
MERCHANTABILITY, TITLE, NON-INFRINGEMENT AND FITNESS FOR A PARTICULAR PURPOSE.
IN NO EVENT SHALL MICROCHIP OR ITS LICENSORS BE LIABLE OR OBLIGATED UNDER
CONTRACT, NEGLIGENCE, STRICT LIABILITY, CONTRIBUTION, BREACH OF WARRANTY, OR
OTHER LEGAL EQUITABLE THEORY ANY DIRECT OR INDIRECT DAMAGES OR EXPENSES
INCLUDING BUT NOT LIMITED TO ANY INCIDENTAL, SPECIAL, INDIRECT, PUNITIVE OR
CONSEQUENTIAL DAMAGES, LOST PROFITS OR LOST DATA, COST OF PROCUREMENT OF
SUBSTITUTE GOODS, TECHNOLOGY, SERVICES, OR ANY CLAIMS BY THIRD PARTIES
(INCLUDING BUT NOT LIMITED TO ANY DEFENSE THEREOF), OR OTHER SIMILAR COSTS.
 *******************************************************************************/

#include <ctype.h>
#include "tcpip/src/tcpip_private.h"
#include "drv_wifi_scan_helper.h"

DRV_WIFI_SCAN_STATUS g_scanStatus;
DRV_WIFI_SCAN_RESULT g_scanResults[50]; // WF_PRESCAN // may change struct later for memory optimization

void DRV_WIFI_HostScanInitialize(void)
{
    g_scanStatus.scanState = 0;
    g_scanStatus.numScanResults = 0;
    g_scanStatus.displayIdx = 0;
}

uint16_t DRV_WIFI_ScanStart(void)
{
    /* If scan already in progress bail out */
    if (IS_SCAN_IN_PROGRESS(g_scanStatus.scanState))
    {
        return DRV_WIFI_ERROR_OPERATION_CANCELLED;
    }
    if (DRV_WIFI_Scan(true) != DRV_WIFI_SUCCESS)
    {
        return DRV_WIFI_ERROR_OPERATION_CANCELLED;
    }
    SCAN_SET_IN_PROGRESS(g_scanStatus.scanState);

    return DRV_WIFI_SUCCESS;
}

uint16_t DRV_WIFI_HostScanResultGet(uint8_t idx, DRV_WIFI_SCAN_RESULT *p_ScanResult)
{
    if (idx >= g_scanStatus.numScanResults)
    {
        return DRV_WIFI_ERROR_INVALID_PARAM;
    }

    DRV_WIFI_ScanResultGet(idx, p_ScanResult);
    //p_ScanResult->ssid[p_ScanResult->ssidLen] = 0; /* Terminate. */

    return DRV_WIFI_SUCCESS;
}

void DRV_WIFI_HostScanEventHandler(uint16_t scanResults)
{
    /* Clear the scan in progress flag. */
    SCAN_CLEAR_IN_PROGRESS(g_scanStatus.scanState);

    SCAN_SET_VALID(g_scanStatus.scanState);

    SCAN_SET_DISPLAY(g_scanStatus.scanState);
    g_scanStatus.numScanResults = scanResults; /* Cache number of APs found in scan. */
    g_scanStatus.displayIdx = 0;
}

void DRV_WIFI_ScanResultDisplayManager(void)
{
    int i;
    char ssid[32 + 1];
    static bool displayHeader = true;
    DRV_WIFI_SCAN_RESULT scanResult;

    if ((g_scanStatus.numScanResults == 0) ||
        (!IS_SCAN_STATE_DISPLAY(g_scanStatus.scanState)) ||
        (IS_SCAN_IN_PROGRESS(g_scanStatus.scanState)) ||
        (!IS_SCAN_STATE_VALID(g_scanStatus.scanState)))
    {
       SCAN_CLEAR_DISPLAY(g_scanStatus.scanState);
       return;
    }

    if (displayHeader) {
        g_scanStatus.displayIdx = 0;
        SYS_CONSOLE_MESSAGE((const char *)"\nWi-Fi scan results ... \r\n\n");
        if (g_scanStatus.numScanResults == 0) {
            /* 
             * No AP found, re-initialize the scan context state. Scan is not allowed in Soft AP mode.
             * However, g_scanStatus.numScanResults needs to be preserved for webpage using.
             */
            SYS_CONSOLE_MESSAGE((const char *)"No AP found.\r\n ");
            SCAN_CLEAR_DISPLAY(g_scanStatus.scanState);
            if (WF_CONFIG_PARAMS(networkType) == DRV_WIFI_NETWORK_TYPE_SOFT_AP) {
                SCAN_CLEAR_VALID(g_scanStatus.scanState);
                return;
            } else {
                return;
            }
        } else {
            SYS_CONSOLE_MESSAGE((const char *)"     SSID                             RSSI  Channel\r\n");
            SYS_CONSOLE_MESSAGE((const char *)"     -------------------------------- ----  -------\r\n");
            displayHeader = false;
        }
    }
    
    DRV_WIFI_HostScanResultGet(g_scanStatus.displayIdx, &scanResult);

    memset(ssid, ' ', sizeof(ssid));

    /* Display SSID */
    for (i = 0; i < scanResult.ssidLen; i++)
    {
        if (!isprint(scanResult.ssid[i]))
        {
            ssid[i] = '*';
        }
        else
        {
            ssid[i] = scanResult.ssid[i];
        }
    }

    ssid[32] = 0;

    /* Display Index & SSID */
    //SYS_CONSOLE_PRINT("%2d)  ", g_scanStatus.displayIdx + 1);
    //SYS_CONSOLE_PRINT("%s", ssid);

    /* Display RSSI & Channel */
    /* RSSI_MAX : 128, RSSI_MIN : 43 */
    //SYS_CONSOLE_PRINT(" %2u    %u\r\n", scanResult.rssi, scanResult.channel);

    SYS_CONSOLE_PRINT(" %2d) %s %2u    %u\r\n", g_scanStatus.displayIdx + 1, ssid, scanResult.rssi, scanResult.channel);

    if (WF_CONFIG_PARAMS(networkType) == DRV_WIFI_NETWORK_TYPE_SOFT_AP) // WF_PRESCAN
    {
        g_scanResults[g_scanStatus.displayIdx] = scanResult; // WF_PRESCAN
        if (g_scanStatus.displayIdx == sizeof(g_scanResults) / sizeof(g_scanResults[0]) - 1)
        {
            /* Prescan is finished, re-initialize the scan context state. Scan is not allowed in Soft AP mode.
             * However, g_scanStatus.numScanResults needs to be preserved for webpage using.
             */
            SCAN_CLEAR_VALID(g_scanStatus.scanState);
            SCAN_CLEAR_DISPLAY(g_scanStatus.scanState);
            g_scanStatus.displayIdx = 0;
            displayHeader = true;
        }
    }
    else
    {
        #if (DRV_WIFI_DEFAULT_ADHOC_PRESCAN == DRV_WIFI_ENABLED)
        g_scanResults[g_scanStatus.displayIdx] = scanResult; // WF_PRESCAN
        if (g_scanStatus.displayIdx == sizeof(g_scanResults) / sizeof(g_scanResults[0]) - 1)
        {
            SCAN_CLEAR_DISPLAY(g_scanStatus.scanState);
            g_scanStatus.displayIdx = 0;
            displayHeader = true;
        }
        #endif
    }

    if (++g_scanStatus.displayIdx == g_scanStatus.numScanResults) // Normal scan which does not happen before Wi-Fi connect.
    {
        if (WF_CONFIG_PARAMS(networkType) == DRV_WIFI_NETWORK_TYPE_SOFT_AP)
        {
            /* Scan is finished, re-initialize the scan context state. Scan is not allowed in Soft AP mode.
             * However, g_scanStatus.numScanResults needs to be preserved for webpage using.
             */
            SCAN_CLEAR_VALID(g_scanStatus.scanState);
            SCAN_CLEAR_DISPLAY(g_scanStatus.scanState);
            g_scanStatus.displayIdx = 0;
            displayHeader = true;
        }
        else
        {
            SCAN_CLEAR_DISPLAY(g_scanStatus.scanState);
            g_scanStatus.displayIdx = 0;
            displayHeader = true;
        }
    }
}

//DOM-IGNORE-END
